/* vim: set cindent et ts=4 sw=4 sts=4 fdm=marker:
 *
 * bin2asm, a program to convert binary data to assembler statements.
 *
 * Author:  Bas 'Compyx' Wassink
 * Started: 2007-03-13
 * Updated: 2007-03-14
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "optparse.h"
#include "bin2asm.h"

#define emsg_parser(X) \
    fprintf(stderr, "%s:%d: Error registering option '%c' with parser.\n", \
            __FILE__, __LINE__, (X))

/* {{{ typedefs for unsigned types */
typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;
/* }}} */

/* {{{ prototypes for functions used */
void display_help(void);
void display_help_extra(void);
void display_version(void);
void print_bin(uint value, uint size);
/* }}} */

/* {{{ Data section */

/* byte dump statements */
static char DataByte[][6] = {
    ".byte",
    ".byt",
    ".db"
};

/* word dump statements */
static char DataWord[][6] = {
    ".word",
    ".dw"
};

/* struct for holding assembler id strings */
struct asmid_st {
    int  Id;
    char String[6];
};

/* Assembler Id strings */
static struct asmid_st AsmIds[] = {
    { ASM_TURBO, "turbo" },
    { ASM_DREAM, "dream" },
    { ASM_XA65,  "xa65"  },
    { ASM_CA65,  "ca65"  },
    { ASM_KICK,  "kick"  }
};


/* struct for holding assembler syntax defaults */
struct asmdef_st {
    int Byte;      /* default byte statement */
    int Word;      /* default word statement */
    int Cols;      /* default number of columns of data per line */
};

/* default syntax settings per assembler */
static struct asmdef_st AsmDefaults[] = {
    /* Turbo Assembler */
    { DB_BYTE, DB_WORD, 6},
    /* Dream Assembler, CA65, Kick Assembler */
    { DB_BYTE, DB_WORD, 8},
    /* XA65 */
    { DB_BYT,  DB_WORD, 8},
};

/* format specifiers for numbers (struct) */
struct numfmt_st {
    int  Base;
    char Format[4];
};

/* format specifiers for numbers */
static struct numfmt_st NumberFormats[] = {
    /* Binary */
    {  2, "bin" },
    /* Octal */
    {  8, "oct" },
    /* Decimal */
    { 10, "dec" },
    /* Hexadecimal */
    { 16, "hex" }
};

/* }}} */


/* Entry point */
int main(int argc, char *argv[])
{
    /* variables used by optparse */
    char   *filename  = NULL;       /* filename of input data */
    char   *asmstring = NULL;       /* assembler syntax identifier string */
    char   *byteop    = NULL;       /* statement to use for byte data */
    char   *wordop    = NULL;       /* statement to use for word data */
    char   *format    = NULL;       /* number format specifier */
    ulong   columns   = 0;          /* int:  columns of data per line */
    ulong   dumpsize  = 0;          /* number of bytes to dump */
    ulong   skip      = 0;          /* offset in file to start dumping */
    int     parsed    = 0;          /* number of options parsed */
    int     words     = 0;          /* bool: dump words, not bytes */
    /* other variables */
    FILE   *fp;                     /* file pointer for reading data */
    size_t  i;                      /* index for variable operations */
    ulong   col        = 0;         /* current column of output */
    ulong   dumped     = 0;         /* number of bytes already dumped */
    int     statbase   = 16;        /* numeric output format for data */
    int     statbyte   = DB_BYTE;   /* statement for bytes */
    int     statword   = DB_WORD;   /* statement for words */
    int     asmid      = ASM_DREAM; /* default syntax */
    int     found;                  /* flag used for searching through data */
    int     done       = 0;         /* flag used in main loop */
    int     lsb        = 0;         /* lsb of data */
    int     msb        = 0;         /* msb of data */


    /* preliminary check: at least a filename is needed */
    if (argc < 2) {
        printf("I need at least a file name to work with, type `%s --help`"
               " for help.\n", PRG_NAME);
        return EXIT_FAILURE;
    }

    /*
     * Parse command line options and process them
     */

    /* register 'optparse_exit()' with atexit(), so I don't need to call it
     * everytime I encounter an error */
    if (atexit(optparse_exit) != 0) {
        fprintf(stderr,
                "%s:%d: Error registering optparse_exit(), aborted.\n",
                __FILE__, __LINE__);
        return EXIT_FAILURE;
    }

    /* initialize parser */
    optparse_init();

    /* register help and version message functions */
    optparse_help(display_help);
    optparse_help_extra(display_help_extra);
    optparse_version(display_version);

    /* populate parser with boolean variables */
    if (optparse_add(&words, OPT_BOOL, 'W', "dump-words",
                     "Dump data in words, not bytes") == OPT_FALSE) {
        emsg_parser('W');
        return EXIT_FAILURE;
    }

    /* populate parser with numeric options */
    if (optparse_add(&columns, OPT_ULONG, 'c', "columns",
                     "number of columns of data per line") == OPT_FALSE) {
        emsg_parser('c');
        return EXIT_FAILURE;
    }
    if (optparse_add(&skip, OPT_ULONG, 's', "skip",
                     "offset to start dump (default = 0)") == OPT_FALSE) {
        emsg_parser('s');
        return EXIT_FAILURE;
    }
    if (optparse_add(&dumpsize, OPT_ULONG, 'n', "dump-size",
                     "number of bytes to dump (default = all)") == OPT_FALSE) {
        emsg_parser('n');
        return EXIT_FAILURE;
    }

    /* populate parser with string options */
    if (optparse_add(&asmstring, OPT_STRING, 'a', "asm",
                     "Set assembler syntax (default=dream)\n"
                     "('turbo', 'dream', 'kick', 'xa65' or 'ca65')")
            == OPT_FALSE) {
        emsg_parser('a');
        return EXIT_FAILURE;
    }
    if (optparse_add(&byteop, OPT_STRING, 'b', "byte",
                     "Set statement to use for bytes\n"
                     "('.byte', '.byt' or 'db')") == OPT_FALSE) {
        emsg_parser('b');
        return EXIT_FAILURE;
    }
    if (optparse_add(&wordop, OPT_STRING, 'w', "word",
                     "Set statement to use for words\n"
                     "('.word' or '.dw')") == OPT_FALSE) {
        emsg_parser('w');
        return EXIT_FAILURE;
    }
    if (optparse_add(&format, OPT_STRING, 'f', "format",
                     "Specify numeric output format\n"
                     "('bin', 'oct', 'dec' or 'hex')") == OPT_FALSE) {
        emsg_parser('f');
        return EXIT_FAILURE;
    }

    /* execute parser */
    parsed = optparse_exec(argc, argv, &filename);
    if (parsed == OPT_ERROR) {
        fprintf(stderr, "Error parsing command line options, see messages.\n");
        return EXIT_FAILURE;
    } else if ((parsed == OPT_HELP) || (parsed == OPT_VERSION)) {
        return EXIT_SUCCESS;
    }
#ifdef DEBUG
    printf("parsed %d option(s).\n", parsed);
#endif

    /* determine assembler syntax defaults to use */
    if (asmstring == NULL) {
#ifdef DEBUG
        (void)puts("No assembler syntax specified, using default.");
#endif
        asmid = ASM_DREAM;
    } else {
#ifdef DEBUG
        printf("Assembler '%s' specified.\n", asmstring);
#endif
        for (i = 0; i < sizeof AsmIds / sizeof(struct asmid_st); i++) {
            if (strcmp(AsmIds[i].String, asmstring) == 0) {
                asmid = i;
            }
        }
#ifdef DEBUG
        printf("Using AsmId '%s'.\n", AsmIds[asmid].String);
#endif
    }

    /* determine number out format (default already set) */
    if (format != NULL) {
        /* compare specified format with list */
        for (i = 0; i < sizeof NumberFormats / sizeof NumberFormats[0]; i++) {
            if (strcmp(NumberFormats[i].Format, format) == 0) {
                /* specifier found, set base */
                statbase = NumberFormats[i].Base;
            }
        }
    }

#ifdef DEBUG
    printf("Outputting numbers in base %d\n", statbase);
#endif

    /* check number of columns */
    if (columns == 0) {
        /* nothing specified, use assembler default */
        columns = AsmDefaults[asmid].Cols;
    }
#ifdef DEBUG
    printf("Outputting data in %lu columns\n", columns);
#endif
  
    /* check if byte or word statements have been provided */
    if (byteop != NULL) {
        /* yes, check with available data */
        found = 0;
        for (i = 0; i < sizeof DataByte / sizeof DataByte[0]; i++) {
            if (strcmp(DataByte[i], byteop) == 0) {
                /* found it, set statbyte */
                statbyte = i;
                found = 1;
            }
        }
        /* check if end of list was reached */
        if (found == 0) {
            /* set default from assembler syntax */
            statbyte = AsmDefaults[asmid].Byte;
        }
    } else {
        /* set statbyte from assembler defaults */
        statbyte = AsmDefaults[asmid].Byte;
    }
    if (wordop != NULL) {
        /* yes, check with available data */
        found = 0;
        for (i = 0; i < sizeof DataWord / sizeof DataWord[0]; i++) {
            if (strcmp(DataWord[i], wordop) == 0) {
                /* found it, set statbyte */
                statword = i;
                found = 1;
            }
        }
        /* check if end of list was reached */
        if (found == 0) {
            /* set default from assembler syntax */
            statword = AsmDefaults[asmid].Word;
        }
    } else {
        /* set statbyte from assembler defaults */
        statword = AsmDefaults[asmid].Word;
    }

#ifdef DEBUG
    printf("Using '%s' for byte data.\n", DataByte[statbyte]);
    printf("Using '%s' for word data.\n", DataWord[statword]);
#endif

    if (filename == NULL) {
        fprintf(stderr, "Error: no filename specified, aborted.\n");
        return EXIT_FAILURE;
    }
    /* open file for reading */
    fp = fopen(filename, "rb");
    if (fp == NULL) {
        fprintf(stderr, "Couldn't open '%s' for reading.\n", filename);
        perror("OS-message");
        return EXIT_FAILURE;
    }
#ifdef DEBUG
    printf("Opened '%s' for reading.\n", filename);
    if (words != 0) {
        (void)puts("Dumping data as words.");
    } else {
        (void)puts("Dumping data as bytes.");
    }
#endif
    
    /* do I need to skip bytes? */
    if (skip > 0) {
#ifdef DEBUG
        printf("Skipping ahead %lu bytes\n", skip);
#endif
        if (fseek(fp, (long)skip, SEEK_SET) != 0) {
            perror("Error skipping bytes");
            fclose(fp);
        }
    }

    /* main loop */
    while ((done == 0) && ((dumpsize == 0) || ((dumpsize > 0) && (dumped < dumpsize)))) {
        if (col == columns) {
            col = 0;
            printf("\n");
        }
      /* get data from file */
        lsb = fgetc(fp);
        if (lsb == EOF) {
            done = 1;
            msb = 0;
        }
        if ((lsb != EOF) && (words != 0)) {
            /* get msb */
            msb = fgetc(fp);
            if (msb == EOF) {
                done = 1;
                msb = 0;
            }
        }
        if ((done == 0) && (col > 0) && (col < columns)) {
            printf(", ");
        }
        if ((done == 0) && (col == 0)) {
            /* print data statement */
            if (words == 0) {
                printf("%s ", DataByte[statbyte]);
            } else {
                printf("%s ", DataWord[statword]);
            }
        }
        if (done == 0) {
            /* print data */
            switch (statbase) {
            case 2:
                printf("%%");
                if (words == 0) {
                    /* byte */
                    print_bin((uint)lsb, 8);
                } else {
                    print_bin((uint)lsb + (uint)msb * 256, 16);
                }
                break;
            case 8:
                if (words == 0) {
                    printf("&%03o", (uint)lsb);
                } else {
                    printf("&%06o", (uint)lsb + (uint)msb * 256);
                }
                break;
            case 10:
                if (words == 0) {
                    printf("%3u", (uint)lsb);
                } else {
                    printf("%5u", (uint)lsb + (uint)msb * 256);
                }
                break;
            case 16:
                if (words == 0) {
                    printf("$%02x", (uint)lsb);
                } else {
                    printf("$%04x", (uint)lsb + (uint)msb * 256);
                }
                break;
            default:
                break;
            }
        }
        col++;
        if (words == 0) {
            dumped++;
        } else {
            dumped += 2;
        }
    }
    printf("\n");

    /* close file */
    fclose(fp);
    return EXIT_SUCCESS;
}


/* {{{ void display_help(void): help message */
void display_help(void)
{
    printf("%s %s -- help\n", PRG_NAME, PRG_VERSION);
    printf("Usage `%s [options] filename`\n\n", PRG_NAME);
    printf("This program reads binary data from a file and outputs the data in a format\n");
    printf("suitable for a variety of assemblers for 65xx-based systems (such as the C64).\n");
    printf("All output is fed to stdout, so redirect if needed.\n\n");
    printf("List of options:\n\n");

}
/* }}} */

/* {{{ void display_help_extra(void): extra info after option list */
void display_help_extra(void)
{
    printf("\n");
    printf("Assembler defaults:\n");
    printf("Dreamass, Kickassembler and CA65 all output with either '.byte' or '.word',\n");
    printf("with 8 columns of data per line. Turbo Assembler outputs 6 columns of data so\n");
    printf("the generated source can be used in original C64-version.\n");
    printf("XA65 outputs bytes with '.byt', again in 8 columns.\n");
}
/* }}} */

/* {{{ void display_version(void): program name, version and author */
void display_version(void)
{
    printf("%s %s, written by Bas 'Compyx' Wassink\n", PRG_NAME, PRG_VERSION);
}
/* }}} */

/* {{{ void print_bin(uint value, uint size): print binary string, since printf and family have no binary specifier */
void print_bin(uint value, uint size)
{
    while (size > 0) {
        if ((value & (1 << (size - 1))) == 0) {
            printf("0");
        } else {
            printf("1");
        }
        size--;
    }
}
/* }}} */



