/* vim: set cindent et ts=4 sw=4 sts=4 fdm=marker:
 *
 * cli option parser, written by Bas 'Compyx' Wassink (compyx@gmail.com)
 *
 * started: 2006-08-06
 * updated: 2007-03-13 
 *
 * 2007-03-13:  * Added newline posibility in option descriptions.
 * 2007-03-16:  * Cleaned up source and added vim autocommands and folding.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#include "optparse.h"

/* offset for description strings in option list when printing to stdout */
#define DESC_OFFSET 32

/* {{{ Error message macros */

/* malloc error message */
#define opt_error_malloc(BYTES) \
    fprintf(stderr, "%s:%d: failed to allocate %lu bytes of storage\n", \
            __FILE__, __LINE__, (unsigned long)(BYTES))

/* overflow error message */
#define opt_error_overflow(ARG) \
    fprintf(stderr, "%s:%d: overflow error converting arg '%s'\n", \
            __FILE__, __LINE__, ARG)

/* no digits error message */
#define opt_error_digits(ARGUMENT) \
    fprintf(stderr, "%s:%d: error: no digits found in arg '%s'\n", \
            __FILE__, __LINE__, ARGUMENT)

/* no argument after option */
#define opt_error_argument(OPTION) \
    fprintf (stderr, "%s:%d: error: missing argument for option '%s'\n", \
            __FILE__, __LINE__, OPTION)
/* }}} */


/* struct used for building options stack */
struct st_option {
    void             *var;
    int               type;
    int               opt_short;
    const char       *opt_long;
    const char       *desc;
    struct st_option *next;
};

/* root of list of options */
static struct st_option *options;

/* pointers to functions displaying application's help and version info */
static void (*func_help)(void);
static void (*func_help_extra)(void);
static void (*func_version)(void);

/* type descriptors of the options */
static const char *opt_types[] = {
    "BOOL",
    "INT",
    "LONG",
    "ULONG",
    "DOUBLE",
    "STRING"
};

/* {{{ static void              opt_print_options(void)
 * Print options list for use with help function */
static void opt_print_options(void)
{
    struct st_option *node;
    int               n;
    size_t            i;

    /* display help and version options, if specified */
    if (func_help != NULL) {
        n = printf("  --help");
        while (n < DESC_OFFSET) {
            printf(" ");
            n++;
        }
        printf("show help and exit\n");
    }
    if (func_version != NULL) {
        n = printf("  --version");
        while (n < DESC_OFFSET) {
            printf(" ");
            n++;
        }
        printf("show program version and exit\n");
    }
    for (node = options; node != NULL; node = node->next) {
        /* check which options are defined */
        if ((node->opt_short != OPT_NO_SHORT) && (node->opt_long != NULL)) {
            /* print both short and long options */
            if (node->type == OPT_BOOL) {
                n = printf("  -%c, --%s", node->opt_short, node->opt_long);
            } else {
                n = printf("  -%c %s, --%s %s",
                           node->opt_short, opt_types[node->type],
                           node->opt_long, opt_types[node->type]);
            }
        } else if (node->opt_short != OPT_NO_SHORT) {
            /* print short opt only */
            if (node->type == OPT_BOOL) {
                n = printf("  -%c", node->opt_short);
            } else {
                n = printf("  -%c %s", node->opt_short, opt_types[node->type]);
            }
        } else {
            /* print long opt only */
            if (node->type == OPT_BOOL) {
                n = printf("  --%s", node->opt_long);
            } else {
                n = printf("  --%s %s", node->opt_long, opt_types[node->type]);
            }
        }
        /* print description string when present */
        if (node->desc != NULL) {
            if (n > DESC_OFFSET) {
                /* previous text too long, print newline */
                printf("\n");
                n = 0;
            }
            /* add spacing */
            while (n < DESC_OFFSET) {
                printf(" ");
                n++;
            }
            /* parse description, translating newlines to newline + offset */
            i = 0;
            while (i < strlen(node->desc)) {
                if (node->desc[i] == '\n') {
                    printf("\n");
                    for (n = 0; n < DESC_OFFSET; n++) {
                        printf(" ");
                    }
                } else {
                    printf("%c", node->desc[i]);
                }
                i++;
            }
            printf("\n");
        }
    }
}
/* }}} */

/* {{{ static int               opt_get_value(struct st_option *node, int argc, char *argv[], size_t offset)
 * Try to get value from argv */
static int opt_get_value(struct st_option *node, int argc, char *argv[],
                         size_t offset)
{

    long           l;           /* holds result of strtol */
    unsigned long  u;           /* holds result of strtoul */
    double         d;           /* holds result of strtod */
    char          *endptr;      /* holds pointer to one past last used char
                                   from call to strtod */

    int            delta = 0;   /* signal wether an argument after the
                                   option was consumed */

    /* determine type of option */
    switch (node->type) {
    
    case OPT_BOOL :
        /* boolean value, set var to 1 */
        *(int*)node->var = 1;
        break;
       
    case OPT_INT :
        /* int value, convert next argument to int */
        if (offset == (size_t)(argc - 1)) {
            /* no more arguments */
            opt_error_argument(argv[offset]);
            return OPT_ERROR;
        }
        /* try to convert the next argument to long (and then to int) */
        errno = 0;
        /* determine base of number */
        if ((strlen(argv[offset+1]) > 2) && (argv[offset+1][0] == '0')
         && ((argv[offset+1][1] == 'x') || (argv[offset+1][1] == 'X'))) {
            /* base 16 */
            l = strtol(argv[offset+1]+2, &endptr, 16);
        } else {
            l = strtol(argv[offset+1], &endptr, 10);
        }
        /* check for overflow */
        if (errno == ERANGE) {
            opt_error_overflow(argv[offset+1]);
            return OPT_ERROR;
        }
        /* check for invalid argument */
        if (argv[offset+1] == endptr) {
            opt_error_digits(argv[offset+1]);
            return OPT_ERROR;
        }
        /* check for overflow when converting long to int (but only if
         * INT_LONG > INT_MAX) */
#if LONG_MAX > INT_MAX
        if ((l < INT_MIN) || (l > INT_MAX)) {
            opt_error_overflow(argv[offset+1]);
            return OPT_ERROR;
        }
#endif
        /* set value */
        *(int*)node->var = (int)l;
        delta = 1;
        break;
        
    case OPT_LONG :
        /* long value, convert next argument to long */
        if (offset == (size_t)(argc - 1)) {
            /* no more arguments */
            opt_error_argument(argv[offset]);
            return OPT_ERROR;
        }
        /* try to convert the next argument to long */
        errno = 0;
        /* determine base of number */
        if ((strlen(argv[offset+1]) > 2) && (argv[offset+1][0] == '0')
            && ((argv[offset+1][1] == 'x') || (argv[offset+1][1] == 'X'))) {
            /* base 16 */
            l = strtol(argv[offset+1]+2, &endptr, 16);
        } else {
            l = strtol(argv[offset+1], &endptr, 10);
        }
        if (errno == ERANGE) {
            opt_error_overflow(argv[offset+1]);
            return OPT_ERROR;
        }
        /* check for no digits present in argument */
        if (argv[offset+1] == endptr) {
            opt_error_digits(argv[offset+1]);
            return OPT_ERROR;
        }
        /* set value */
        *(long*)node->var = l;
        delta = 1;
        break;
    
    case OPT_ULONG :
        /* unsigned long value, convert next argument to unsigned long */
        if (offset == (size_t)(argc - 1)) {
            /* no more arguments */
            opt_error_argument(argv[offset]);
            return OPT_ERROR;
        }
        /* try to convert the next argument to unsigned long */
        errno = 0;
        /* determine base of number */
        if ((strlen(argv[offset+1]) > 2) && (argv[offset+1][0] == '0')
            && ((argv[offset+1][1] == 'x') || (argv[offset+1][1] == 'X'))) {
            /* base 16 */
            u = strtoul(argv[offset+1]+2, &endptr, 16);
        } else {
            u = strtoul(argv[offset+1], &endptr, 10);
        }
        if (errno == ERANGE) {
            opt_error_overflow(argv[offset+1]);
            return OPT_ERROR;
        }
        if (argv[offset+1] == endptr) {
            opt_error_digits(argv[offset+1]);
            return OPT_ERROR;
        }
        /* set value */
        *(unsigned long*)node->var = u;
        delta = 1;
        break;

    case OPT_DOUBLE :
        /* double value, convert next argument to double */
        if (offset == (size_t)(argc - 1)) {
            /* no more arguments */
            opt_error_argument(argv[offset]);
            return OPT_ERROR;
        }
        /* try to convert the next argument to int */
        errno = 0;
        d = strtod(argv[offset+1], &endptr);
        if (errno == ERANGE) {
            opt_error_overflow(argv[offset+1]);
            return OPT_ERROR;
        }
        if (argv[offset+1] == endptr) {
            opt_error_digits(argv[offset+1]);
            return OPT_ERROR;
        }
        /* set value */
        *(double*)node->var = d;
        delta = 1;
        break;

    case OPT_STRING:
        /* string value, return pointer to argument */
        if (offset == (size_t)(argc - 1)) {
            /* no more arguments */
            opt_error_argument(argv[offset]);
            return OPT_ERROR;
        }
        *(char**)node->var = argv[offset+1];
        delta = 1;
        break;
    default:
        break;
    }

    return delta;
}
/* }}} */

/* {{{ static struct st_option *find_long_opt(const char *opt)
 * Return pointer to option list element or NULL when not found */
static struct st_option *find_long_opt(const char *opt)
{
    struct st_option *node;

    for (node = options; node != NULL; node = node->next) {
        if (strcmp(node->opt_long, opt) == 0) {
            return node;
        }
    }
    return NULL;
}
/* }}} */

/* {{{ static struct st_option *find_short_opt(int opt)
 * Return pointer to option list element or NULL when not found */
static struct st_option *find_short_opt(int opt)
{
    struct st_option *node;

    for (node = options; node != NULL; node = node->next) {
        if (node->opt_short == opt) {
            return node;
        }
    }
    return NULL;
}
/* }}} */


/* Global functions */

/* {{{ void optparse_init(void)
 * Initialize option parser */
void optparse_init(void)
{
    /* initialize list of options */
    options = NULL;
    /* set help function pointer to NULL */
    func_help = NULL;
    /* set version info function pointer to NULL */
    func_version = NULL;
}
/* }}} */

/* {{{ void optparse_help(void (*func)(void))
 * Register help function */
void optparse_help(void (*func)(void))
{
    func_help = func;
}
/* }}} */

/* {{{ void optparse_version(void (*func)(void))
 * Register version function */
void optparse_version(void (*func)(void))
{
    func_version = func;
}
/* }}} */

/* {{{ void optparse_help_extra(void (*func)(void))
 * Register help_extra function (displayed after options list) */
void optparse_help_extra(void (*func)(void))
{
    func_help_extra = func;
}
/* }}} */

/* {{{ int  optparse_add(void *var, int type, int opt_s, const char *opt_l, const char *desc)
 * Add option to the parser */
int optparse_add(void *var, int type, int opt_s, const char *opt_l,
                 const char *desc)
{
    struct st_option *node;

    /* test arguments */
    if (var == NULL) {
        fprintf(stderr, "%s:%d: error: NULL passed as variable pointer\n",
                __FILE__, __LINE__);
        return OPT_FALSE;
    }
    
    /* test type value */
    if ((type < OPT_BOOL) || (type > OPT_STRING)) {
        fprintf(stderr, "%s:%d: specified type of option is not valid\n",
                __FILE__, __LINE__);
        return OPT_FALSE;
    }

    /* allocate storage for option struct */
    node = malloc(sizeof *node);
    if (node == NULL) {
        opt_error_malloc(sizeof *node);
        return OPT_FALSE;
    }

    /* store data in option struct */
    node->var       = var;
    node->type      = type;
    node->opt_short = opt_s;
    /* remove leading dashes from long-opt */
    while (*opt_l == '-') {
        opt_l++;
    }
    node->opt_long = opt_l;
    node->desc     = desc;

    /* add to options list */
    node->next = options;
    options    = node;
    
    return OPT_TRUE;
}
/* }}} */

/* {{{ void optparse_exit(void)
 * Free memory used for optparse options list */
void optparse_exit(void)
{
    struct st_option *next;

    /* iterate through options list, deallocating memory for each entry */
    while (options != NULL) {
        next = options->next;
        free(options);
        options = next;
    }
    /* set list root to NULL to avoid UB with subsequent calls */
    options = NULL;
}
/* }}} */

/* {{{ int  optparse_exec(int argc, char *argv[], char **argp)
 * Run the option parser, returns number of options parsed, or a negative
 * value on error (see optparse.c for defines of these values) */
int optparse_exec(int argc, char *argv[], char **argp)
{
    int               parsed = 0;
    int               d;
    size_t            x;
    size_t            y;
    struct st_option *node;

    if ((argp == NULL) || (argv == NULL) || (argc < 1)) {
        return OPT_ERROR;
    }

    /* check if argc == 1 --> no options to parse */
    if (argc == 1) {
        return 0;
    }

    /* parse options */
    for (x = 1; x < (size_t)argc; x++) {
        if (( argv[x][0] == '-') && (argv[x][1] == '-')) {
            
            /* parse long-opts */

            /* first check for --help */
            if (strcmp(argv[x], "--help") == 0) {
                if (func_help != NULL) {
                    /* call user defined help function */
                    func_help();
                    /* display options list */
                    opt_print_options();
                    /* print extra help info, if any */
                    if (func_help_extra != NULL) {
                        func_help_extra();
                    }
                }
                return OPT_HELP;
            }
            /* check for --version */
            if (strcmp(argv[x], "--version") == 0) {
                if (func_version != NULL) {
                    func_version();
                }
                return OPT_VERSION;
            }
            
            /* check long opt against option list */
            node = find_long_opt(argv[x]+2);
            
            if (node != NULL) {
                /* found long option */
                /* determine value for option */
                d = opt_get_value(node, argc, argv, x);
                if (d < 0) {
                    /* error, message has already been printed, exit */
                    return OPT_ERROR;
                } else {
                    /* add 0 or 1 to x */
                    x += d;
                }
                /* increment parse count */
                parsed++;
            } else {
                fprintf(stderr, "%s%d: error: unknown long option '%s'\n",
                        __FILE__, __LINE__, argv[x]);
                return OPT_ERROR;
            }

        } else if (argv[x][0] == '-') {
            
            /* parse short-opts */
            d = 0;  /* set delta to 0, when delta becomes 1 and the parsed
                       short-opt isn't the last one execution is halted and
                       an error message is printed to stderr */
            for (y = 1; (d == 0) && (y < strlen(argv[x])); y++) {
                /* try to find option in list */
                node = find_short_opt(argv[x][y]);
                if (node != NULL) {
                    /* determine value for option */
                    d = opt_get_value(node, argc, argv, x);
                    if (d < 0) {
                        /* error (message already printed to stderr) */
                        return OPT_ERROR; 
                    } else {
                        /* add 0 or 1 to x */
                        x += d;
                    }

                    /* check for trailing short opts when delta == 1 */
                    if (( d == 1) && (argv[x-1][y+1] != '\0')) {
                        fprintf(stderr, "%s:%d: error: trailing short-opt(s)"
                                "after option '%c'!\n",
                                __FILE__, __LINE__, argv[x-1][y]);
                        return OPT_ERROR;
                    }

                    parsed++;
                } else {
                    /* option(s) remain in current arg, report error and halt */
                    fprintf(stderr,
                            "%s:%d: error: unknown short option '%c'\n",
                            __FILE__, __LINE__, argv[x][y]);
                    return OPT_ERROR;
                }
            }

        } else {
            /* not an option, set argp */
            *argp = argv[x];
        }
    }

    /* return number of retrieved options */
    return parsed;
}
/* }}} */


