/*
 * Header file for optparse.c, a translation unit for parsing cli options.
 *
 * Written by Bas Wassink (compyx@gmail.com)
 *
 * Started: 2006-08-06
 * Updated: 2007-03-13 (added optparse_help_extra and \n in descriptions)
 */

/* boolean values for return values*/
#define OPT_FALSE 0
#define OPT_TRUE  1

/* constants for passing to function calls */
#define OPT_NO_SHORT -1
#define OPT_NO_LONG  NULL

/* special return values */
#define OPT_HELP    -3
#define OPT_VERSION -2
#define OPT_ERROR   -1

/* available types for parsing */
enum op_types {
    OPT_BOOL,   /* boolean, store 0 (false) or 1 (true) in var */
    OPT_INT,    /* int, convert next arg to int */
    OPT_LONG,   /* long, convert next arg to long */
    OPT_ULONG,  /* unsigned long, convert next arg to unsigned long */
    OPT_DOUBLE, /* double, convert next arg to double */
    OPT_STRING  /* string, store pointer to next arg in var */
};


/*
 * Function prototypes, all comments describe the function immediately
 * following the description.
 */

/* Initialize parser (always call this before any other calls!) */
void optparse_init(void);


/* Set function to call when --help is given as an option, after the function
 * is called the parser will output a list of options with their descriptions
 * and their argument types, if any. */
void optparse_help(void (*func)(void));

/* Set function to call when --help is given as an options, displayed after
 * the options list */
void optparse_help_extra(void (*func)(void));

/* Set function to call when --version is given as an option */
void optparse_version(void (*func)(void));


/* Add an option to the parser, argument descriptions:
 *   var     pointer to the variable in which to store the result of the
 *             option or it's argument
 *   type    type of the variable 'var', which is one of:
 *             OPT_BOOL    OPT_TRUE is stored in 'var' when option encountered
 *             OPT_INT     option takes int argument
 *             OPT_LONG    option takes long argument
 *             OPT_ULONG   option takes unsigned long argument
 *             OPT_DOUBLE  option takes double argument
 *             OPT_STRING  pointer to argument after option is stored in 'var'
 *   opt_s   short option identifier, single alphanumberic character.
 *   opt_l   long option identifier, string
 *   desc    description of option and/or argument (string)
 *
 * Return values: OPT_TRUE on success, OPT_FALSE on error.
 *
 *
 * 2007-01-05:  The parser now supports base16 numbers for int, long and
 *              unsigned long by using a prefix of '0x' or '0X'.
 */
int optparse_add(void *var, int type,
                 int opt_s, const char *opt_l, const char *desc);


/* Run parser, arguments:
 *   argc      the value of argc from the call to main
 *   argv      the value of argv from the call to main
 *   argp      pointer to pointer to char
 *
 * Return value:
 *   The function returns a count of options encountered (0 - INT_MAX) or
 *   OPT_ERROR on error, OPT_HELP when '--help' was encountered, or
 *   OPT_VERSION when '--version' was encountered.
 *
 *   argp holds a pointer to the first non-option argument in argv
 */
int optparse_exec(int argc, char *arv[], char **arpg);


/* Clean up all memory used by the parser (always call this when done!) */ 
void optparse_exit(void);

